/* $Id: sofi_spc_arc.c,v 1.51 2013-03-12 08:04:53 llundin Exp $
 *
 * This file is part of the SOFI Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:04:53 $
 * $Revision: 1.51 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <math.h>
#include <float.h>
#include <cpl.h>

#include "irplib_plugin.h"
#include "irplib_utils.h"
#include "irplib_distortion.h"
#include "irplib_wlxcorr.h"

#include "sofi_utils.h"
#include "sofi_wavelength.h"
#include "sofi_pfits.h"
#include "sofi_dfs.h"

/*-----------------------------------------------------------------------------
                            Defines
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "sofi_spc_arc"

#define SOFI_ARC_SATURATION    20000

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static cpl_error_code sofi_spc_arc_reduce(cpl_frameset *,
                                          const cpl_parameterlist *, 
                                          const char *, const char *,
                                          const cpl_frameset *,
                                          cpl_errorstate);
static cpl_table * sofi_spc_arc_compute(const cpl_image *, const char *,
        const char *, const char *, const char *, cpl_table **, cpl_image **);
static cpl_error_code sofi_spc_arc_save(cpl_frameset *, const cpl_table *,
                                        const cpl_table *,
                                        const cpl_image *, const char *,
                                        const cpl_frameset *,
                                        const cpl_parameterlist *);
static int sofi_spc_arc_compare(const cpl_frame *, const cpl_frame *);
static int * sofi_spc_arc_find_lamps(const cpl_frameset *);
static int sofi_is_xenon_lamp_active(const cpl_propertylist *);
static int sofi_is_neon_lamp_active(const cpl_propertylist *);
 
cpl_recipe_define(sofi_spc_arc, SOFI_BINARY_VERSION,
                  "Lars Lundin", PACKAGE_BUGREPORT, "2002,2003,2009", 
                  "SOFI Spectro arc recipe",
                  RECIPE_STRING " -- SOFI Spectro arc recipe.\n"
                  "The files listed in the Set Of Frames (sof-file) " 
                  "must be tagged:\n"
                  "raw-file.fits "SOFI_SPC_ARC_RAW" or\n"
                  "xe-cat.fits "SOFI_CALPRO_XE_CAT" or\n"
                  "ne-cat.fits "SOFI_CALPRO_NE_CAT"\n");

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    int             display;
    int             rej_left;
    int             rej_right;
    int             subdark;
    int             arc_max_w;
    int             out_corr;
    cpl_size        degree;
    double          wl_err;
    int             wl_nsamples;
    int             ppm;
    const char  *   lines;
    /* Outputs */
    int             set_nb;
    int             frame_nb;
    int             nb_saturated;
    double          disprel_cc;
    double          fwhm_med;
    int             fwhm_good;
} sofi_spc_arc_config;

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Fill the recipe parameterlist
  @param    self  The parameterlist
  @return   0 iff everything is ok

 */
/*----------------------------------------------------------------------------*/
static
cpl_error_code sofi_spc_arc_fill_parameterlist(cpl_parameterlist * self)
{
    const char * context = PACKAGE "." RECIPE_STRING;
    cpl_error_code err;

    cpl_ensure_code(self, CPL_ERROR_NULL_INPUT);

    /* Fill the parameters list */

    /* --rejected */
    err = irplib_parameterlist_set_string(self, PACKAGE, RECIPE_STRING,
                                          "rejected", "100,100", "rej", context,
                                          "Left, right rejections [pixel]");
    cpl_ensure_code(!err, err);

    /* --subdark */
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING, "subdark",
                                        CPL_FALSE, NULL, context,
                                        "Enable dark subtraction");
    cpl_ensure_code(!err, err);

    /* --arc_max_w */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING, "arc_max_w",
                                       33, NULL, context, "Maximum supported "
                                       "arc width [pixel]");
    cpl_ensure_code(!err, err);

    /* --out_corr */
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING, "out_corr",
                                        CPL_FALSE, NULL, context,
                                        "Enable correction of output images");
    cpl_ensure_code(!err, err);

    /* --display */ 
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING, "display",
                                        CPL_FALSE, NULL, context,
                                        "Enable plotting");
    cpl_ensure_code(!err, err);

    /* --degree */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING, "degree",
                                       2, NULL, context, "Degree of the "
                                       "wavelength dispersion polynomial");
    cpl_ensure_code(!err, err);

    /* --wl_nsamples */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "wl_nsamples", 100, NULL, context,
                                       "Number of samples for the wavelength "
                                       "calibration");
    cpl_ensure_code(!err, err);

    /* --wl_err */
    err = irplib_parameterlist_set_double(self, PACKAGE, RECIPE_STRING,
                                          "wl_err", 1000.0, NULL, context,
                                          "The wavelength error [Angstrom]");
    cpl_ensure_code(!err, err);

    /* --lines */
    err = irplib_parameterlist_set_string(self, PACKAGE, RECIPE_STRING,
                                          "lines", NULL, NULL, context,
                                          "Lines ASCII-file");
    cpl_ensure_code(!err, err);

    /* --ppm */
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING, "ppm",
                                        CPL_FALSE, NULL, context,
                                        "Enable Point Pattern Matching");
    cpl_ensure_code(!err, err);

    return CPL_ERROR_NONE;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int sofi_spc_arc(cpl_frameset            * framelist,
                        const cpl_parameterlist * parlist)
{
    cpl_errorstate          prestate = cpl_errorstate_get();
    const char          *   sval;
    cpl_size            *   labels = NULL;
    cpl_size                nlabels;
    const int               nframes = cpl_frameset_get_size(framelist);
    int                     narcs;
    const char          *   xe;
    const char          *   ne;
    cpl_frameset        *   arcframes = NULL;
    cpl_frameset        *   arc_one = NULL;
    int                     i;
    
    /* Initialise */
    sofi_spc_arc_config.lines = NULL;

    /* Retrieve input parameters */

    /* Rejection parameters */
    sval = irplib_parameterlist_get_string(parlist, PACKAGE, RECIPE_STRING,
                                           "rejected");
    bug_if(sval == NULL);

    skip_if (sscanf(sval, "%d,%d",
                    &sofi_spc_arc_config.rej_left,
                    &sofi_spc_arc_config.rej_right) != 2);

    /* Output corrected images */
    sofi_spc_arc_config.out_corr
        = irplib_parameterlist_get_bool(parlist, PACKAGE, RECIPE_STRING,
                                        "out_corr");
    /* Dark subtraction */
    sofi_spc_arc_config.subdark
        = irplib_parameterlist_get_bool(parlist, PACKAGE, RECIPE_STRING,
                                        "subdark");
    /* ppm */
    sofi_spc_arc_config.ppm
        = irplib_parameterlist_get_bool(parlist, PACKAGE, RECIPE_STRING,
                                        "ppm");
    /* Polynomial degree */
    sofi_spc_arc_config.degree
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "degree");

    /* Number of samples */
    sofi_spc_arc_config.wl_nsamples
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "wl_nsamples");

    /* Wavelength error factor */
    sofi_spc_arc_config.wl_err
        = irplib_parameterlist_get_double(parlist, PACKAGE, RECIPE_STRING,
                                          "wl_err");

    /* Arc max width in pixels */
    sofi_spc_arc_config.arc_max_w
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "arc_max_w");

    /* Display */
    sofi_spc_arc_config.display
        = irplib_parameterlist_get_bool(parlist, PACKAGE, RECIPE_STRING,
                                        "display");

    /* Lines */
    sofi_spc_arc_config.lines
        = irplib_parameterlist_get_string(parlist, PACKAGE, RECIPE_STRING,
                                          "lines");

    /* Identify the RAW and CALIB frames in the input frameset */
    skip_if (sofi_dfs_set_groups(framelist));

    /* Retrieve raw frames */
    arcframes = sofi_extract_frameset(framelist, SOFI_SPC_ARC_RAW);
    error_if (arcframes == NULL, CPL_ERROR_DATA_NOT_FOUND,
              "Cannot find arc frames in the input frame set");

    narcs = cpl_frameset_get_size(arcframes);

    /* Retrieve calibration data */
    xe = sofi_extract_filename(framelist, SOFI_CALPRO_XE_CAT);
    ne = sofi_extract_filename(framelist, SOFI_CALPRO_NE_CAT);

    /* Labelise all input frames */
    labels = cpl_frameset_labelise(arcframes, sofi_spc_arc_compare, &nlabels);
    error_if (labels == NULL, CPL_ERROR_ILLEGAL_INPUT,
              "Cannot labelise the input frames");
   
    /* Extract settings and reduce each of them */
    for (i=0; i < nlabels; i++) {

        /* Reduce data set nb i */
        cpl_msg_info(cpl_func, "Reducing set %d/%d", i+1, (int)nlabels);

        sofi_spc_arc_config.set_nb = i+1;

        cpl_frameset_delete(arc_one);
        arc_one = cpl_frameset_extract(arcframes, labels, i);

        /* Reduce the set */
        if (sofi_spc_arc_reduce(framelist, parlist, xe, ne, arc_one, prestate))
            (void)cpl_error_set_message(cpl_func, cpl_error_get_code(), "Could "
                                        "not reduce set %d of %d with %d arc "
                                        "frame(s)", i+1, (int)nlabels,
                                        (int)cpl_frameset_get_size(arc_one));
    }

    error_if (cpl_frameset_get_size(framelist) == nframes,
              CPL_ERROR_ILLEGAL_INPUT, "No products produced from %d arc "
              "frames grouped into %d set(s)", narcs, (int)nlabels);

    end_skip;
    
    cpl_frameset_delete(arcframes);
    cpl_frameset_delete(arc_one);
    cpl_free(labels);

    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part for SW mode 
  @param    set_tot     the input frame set
  @param    parlist     the input list of parameters
  @param    xe          the xe lines catalog
  @param    ne          the ne lines catalog
  @param    arcframes   the frames list with arc frames
  @param    prestate    The (clean) CPL errorstate prior to reduction
  @return   0 iff ok
  @note Resets any CPL error(s) prior to reducing arc data
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code sofi_spc_arc_reduce(cpl_frameset            * set_tot,
                                          const cpl_parameterlist * parlist,
                                          const char              * xe,
                                          const char              * ne,
                                          const cpl_frameset      * arcframes,
                                          cpl_errorstate            prestate)
{
    /* Look for each file which lamp is activated */
    /* lamps[i]==0   --->   frame i : lamps off */
    /* lamps[i]==1   --->   frame i : Xenon lamp on */
    /* lamps[i]==2   --->   frame i : Neon lamp on */
    /* lamps[i]==3   --->   frame i : Ne and Xe lamps on  */
    int                 *   lamps = sofi_spc_arc_find_lamps(arcframes);
    const int               nframes = cpl_frameset_get_size(arcframes);
    char                    lines_table[16];
    const cpl_frame     *   cur_frame;
    const char          *   cur_fname;
    cpl_image           *   xenon = NULL;
    cpl_image           *   neon  = NULL;
    cpl_image           *   xe_ne = NULL;
    cpl_image           *   dark  = NULL;
    cpl_image           *   to_compute;
    cpl_table           *   arcs_fwhm;
    cpl_image           *   out_corr;
    cpl_table           *   out_table;
    int                     i;

    /* Initialise */
    sofi_spc_arc_config.frame_nb = 0;
    
    if (lamps == NULL) return cpl_error_set_where(cpl_func);

    /* Get the images */
    for (i=0; i<nframes; i++) {
        cur_frame = cpl_frameset_get_position_const(arcframes, i);
        cur_fname = cpl_frame_get_filename(cur_frame);
        if ((lamps[i] == 0) && (dark == NULL)) {
            cpl_msg_info(cpl_func, "Dark image: [%s]", cur_fname);
            dark = cpl_image_load(cur_fname, CPL_TYPE_FLOAT, 0, 0);
        } else if ((lamps[i] == 1) && (xenon == NULL)) {
            cpl_msg_info(cpl_func, "Xenon lamp: [%s]", cur_fname);
            xenon = cpl_image_load(cur_fname, CPL_TYPE_FLOAT, 0, 0);
        } else if ((lamps[i] == 2) && (neon == NULL)) {
            cpl_msg_info(cpl_func, "Neon lamp: [%s]", cur_fname);
            neon = cpl_image_load(cur_fname, CPL_TYPE_FLOAT, 0, 0);
        } else if ((lamps[i] == 3) && (xe_ne == NULL)) {
            cpl_msg_info(cpl_func, "Xenon+Neon lamp: [%s]", cur_fname);
            xe_ne = cpl_image_load(cur_fname, CPL_TYPE_FLOAT, 0, 0);
        }
    }
    cpl_free(lamps);
   
    /* All lamps are switched off */
    if (!xenon && !neon && !xe_ne) {
        cpl_image_delete(dark);
        return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                                     "Neither xenon nor neon lamp activated");
    }
   
    /* Get the first file name */
    cur_frame = cpl_frameset_get_position_const(arcframes, 0);
    cur_fname = cpl_frame_get_filename(cur_frame);
    
    /* Xenon lines */
    if (xenon) {
        if (!cpl_errorstate_is_equal(prestate)) {
            cpl_msg_warning(cpl_func, "Resetting CPL error before reducing "
                            "Xenon data:");
            cpl_errorstate_dump(prestate, CPL_FALSE, NULL);
            cpl_errorstate_set(prestate);
        }
        if (dark) to_compute = cpl_image_subtract_create(xenon, dark);
        else to_compute = cpl_image_duplicate(xenon);
        cpl_image_delete(xenon);
        strcpy(lines_table, "Xe");

        /* Apply the reduction */
        cpl_msg_info(cpl_func, "Apply the reduction");
        sofi_spc_arc_config.frame_nb ++;
        cpl_msg_indent_more();
        if ((out_table = sofi_spc_arc_compute(to_compute, cur_fname, 
                        lines_table, xe, ne, &arcs_fwhm, &out_corr)) == NULL) {
            cpl_msg_error(cpl_func, "arc reduction computation failed");
            cpl_image_delete(to_compute);
        } else {
            /* Save the products */
            cpl_image_delete(to_compute);
            cpl_msg_info(cpl_func, "Save the products");
            sofi_spc_arc_save(set_tot, out_table, arcs_fwhm, out_corr,
                              lines_table, arcframes, parlist);
            cpl_table_delete(out_table);
            if (arcs_fwhm) cpl_table_delete(arcs_fwhm);
            if (out_corr) cpl_image_delete(out_corr);
        }
        cpl_msg_indent_less();
    }

    /* Neon lines */
    if (neon) {
        if (!cpl_errorstate_is_equal(prestate)) {
            cpl_msg_warning(cpl_func, "Resetting CPL error before reducing "
                            "Neon data:");
            cpl_errorstate_dump(prestate, CPL_FALSE, NULL);
            cpl_errorstate_set(prestate);
        }
        if (dark) to_compute = cpl_image_subtract_create(neon, dark);
        else to_compute = cpl_image_duplicate(neon);
        cpl_image_delete(neon);
        strcpy(lines_table, "Ne");

        /* Apply the reduction */
        cpl_msg_info(cpl_func, "Apply the reduction");
        sofi_spc_arc_config.frame_nb ++;
        cpl_msg_indent_more();
        if ((out_table = sofi_spc_arc_compute(to_compute, cur_fname, 
                        lines_table, xe, ne, &arcs_fwhm, &out_corr)) == NULL) {
            cpl_msg_error(cpl_func, "arc reduction computation failed");
            cpl_image_delete(to_compute);
        } else {
            /* Save the products */
            cpl_image_delete(to_compute);
            cpl_msg_info(cpl_func, "Save the products");
            sofi_spc_arc_save(set_tot, out_table, arcs_fwhm, out_corr,
                              lines_table, arcframes, parlist);
            cpl_table_delete(out_table);
            if (arcs_fwhm) cpl_table_delete(arcs_fwhm);
            if (out_corr) cpl_image_delete(out_corr);
        }
        cpl_msg_indent_less();
    }


    /* Xenon+Neon lines */
    if (xe_ne) {
        if (!cpl_errorstate_is_equal(prestate)) {
            cpl_msg_warning(cpl_func, "Resetting CPL error before reducing "
                            "Xenon+Neon data:");
            cpl_errorstate_dump(prestate, CPL_FALSE, NULL);
            cpl_errorstate_set(prestate);
        }
        if (dark) to_compute = cpl_image_subtract_create(xe_ne, dark);
        else to_compute = cpl_image_duplicate(xe_ne);
        cpl_image_delete(xe_ne);
        strcpy(lines_table, "Xe+Ne");

        /* Apply the reduction */
        cpl_msg_info(cpl_func, "Apply the reduction");
        sofi_spc_arc_config.frame_nb ++;
        cpl_msg_indent_more();
        if ((out_table = sofi_spc_arc_compute(to_compute, cur_fname, 
                        lines_table, xe, ne, &arcs_fwhm, &out_corr)) == NULL) {
            cpl_msg_error(cpl_func, "arc reduction computation failed");
            cpl_image_delete(to_compute);
        } else {
            /* Save the products */
            cpl_image_delete(to_compute);
            cpl_msg_info(cpl_func, "Save the products");
            sofi_spc_arc_save(set_tot, out_table, arcs_fwhm, out_corr,
                              lines_table, arcframes, parlist);
            cpl_table_delete(out_table);
            if (arcs_fwhm) cpl_table_delete(arcs_fwhm);
            if (out_corr) cpl_image_delete(out_corr);
        }
        cpl_msg_indent_less();
    }
    cpl_image_delete(dark);
    
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Determine the distortion, correct, and wl calibrate
  @param    in          input image
  @param    in_name     input file name
  @param    lines_table catalog
  @param    xe          the xe lines catalog
  @param    ne          the ne lines catalog
  @param    arcs_fwhm   table with the computed arcs FHWM
  @param    out_corr    the distortion corrected image
  @return   the result table with 3 columns with the 2d poly and the
  last column with the  1D polynomial describing the dispersion
 */
/*----------------------------------------------------------------------------*/
static cpl_table * sofi_spc_arc_compute(
        const cpl_image     *   in,
        const char          *   in_name,
        const char          *   lines_table,
        const char          *   xe,
        const char          *   ne,
        cpl_table           **  arcs_fwhm,
        cpl_image           **  out_corr)
{
    cpl_image           *   flipped;
    cpl_table           *   out_tab;
    int                     xmin, ymin, xmax, ymax;
    int                     nx, ny;
    cpl_mask            *   saturation_map;
    double                  slit_width;
    cpl_polynomial      *   distor_poly_flipped;
    cpl_polynomial      *   distor_poly;
    cpl_polynomial      *   poly2d_id;
    cpl_vector          *   profile;
    cpl_image           *   in_corr;
    cpl_apertures       *   arcs;
    cpl_polynomial      *   phdisprel;
    cpl_polynomial      *   disprel;
    cpl_size                power[5];
    int                     degree;
    double                  a, b, c, d, e, xc;
    cpl_size                power_flipped[2];
    double                  fwhm_x, fwhm_y;
    cpl_vector          *   fwhm_valid;
    int                     i;

    /* Initialise */
    nx = cpl_image_get_size_x(in);
    ny = cpl_image_get_size_y(in);
    *arcs_fwhm = NULL;
    *out_corr = NULL;

    /* Define zone */
    xmin = sofi_spc_arc_config.rej_left + 1;
    xmax = nx - sofi_spc_arc_config.rej_right;
    ymin = 1;
    ymax = ny;

    /* Compute the number of saturated pixels */
    saturation_map = cpl_mask_threshold_image_create(in,
            SOFI_ARC_SATURATION, DBL_MAX);
    if (saturation_map != NULL) {
        sofi_spc_arc_config.nb_saturated = cpl_mask_count(saturation_map);
        cpl_mask_delete(saturation_map);
    } else {
        sofi_spc_arc_config.nb_saturated = 0;
    }

    /* Distortion estimation */
    cpl_msg_info(cpl_func, "Estimate the distortion");
    cpl_msg_indent_more();
    /* Flip the image */
    flipped = cpl_image_duplicate(in);
    cpl_image_flip(flipped, 1);
    if ((distor_poly_flipped = irplib_distortion_estimate(flipped, 
                    ymin, xmin, ymax, xmax, 
                    sofi_spc_arc_config.subdark, SOFI_ARC_SATURATION,
                    sofi_spc_arc_config.arc_max_w, 0.33,2,&arcs)) == NULL) {
        cpl_msg_error(cpl_func, "cannot estimage distortion");
        cpl_image_delete(flipped);
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_image_delete(flipped);
    cpl_msg_indent_less();
    
    /* Flip the polynomial back */
    distor_poly = cpl_polynomial_duplicate(distor_poly_flipped);
    power[0] = 1; power[1] = 0; power_flipped[0] = 0; power_flipped[1] = 1;
    cpl_polynomial_set_coeff(distor_poly, power, 
            cpl_polynomial_get_coeff(distor_poly_flipped, power_flipped));
    
    power[0] = 2; power[1] = 0; power_flipped[0] = 0; power_flipped[1] = 2;
    cpl_polynomial_set_coeff(distor_poly, power, 
            cpl_polynomial_get_coeff(distor_poly_flipped, power_flipped));

    power[0] = 0; power[1] = 1; power_flipped[0] = 1; power_flipped[1] = 0;
    cpl_polynomial_set_coeff(distor_poly, power, 
            cpl_polynomial_get_coeff(distor_poly_flipped, power_flipped));

    power[0] = 0; power[1] = 2; power_flipped[0] = 2; power_flipped[1] = 0;
    cpl_polynomial_set_coeff(distor_poly, power, 
            cpl_polynomial_get_coeff(distor_poly_flipped, power_flipped));
    cpl_polynomial_delete(distor_poly_flipped);

    /* Distortion correction */
    cpl_msg_info(cpl_func, "Correct the distortion");
    cpl_msg_indent_more();
    in_corr = cpl_image_duplicate(in);

    /* Create the identity polynomial */
    poly2d_id = cpl_polynomial_new(2);
    power[0] = 1; power[1] = 0;
    cpl_polynomial_set_coeff(poly2d_id, power, 1.0);
    
    /* Create the kernel */
    profile = cpl_vector_new(CPL_KERNEL_DEF_SAMPLES);
    cpl_vector_fill_kernel_profile(profile, CPL_KERNEL_DEFAULT,
            CPL_KERNEL_DEF_WIDTH);

    /* Apply the distortion */
    if (cpl_image_warp_polynomial(in_corr, in, poly2d_id, distor_poly,
            profile, CPL_KERNEL_DEF_WIDTH, profile,
            CPL_KERNEL_DEF_WIDTH) != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "cannot correct the distortion");
        cpl_image_delete(in_corr);
        cpl_polynomial_delete(poly2d_id);
        cpl_polynomial_delete(distor_poly);
        cpl_vector_delete(profile);
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_polynomial_delete(poly2d_id);
    cpl_vector_delete(profile);
    cpl_msg_indent_less();

    /* Compute the FWHM of the detected arcs */
    cpl_msg_info(cpl_func, "Compute the FWHM of the detected arcs");
    /* Create the fwhms table */
    *arcs_fwhm = cpl_table_new(cpl_apertures_get_size(arcs));
    cpl_table_new_column(*arcs_fwhm, "POS", CPL_TYPE_DOUBLE);
    cpl_table_new_column(*arcs_fwhm, "FWHM", CPL_TYPE_DOUBLE);
    cpl_table_new_column(*arcs_fwhm, "FLUX", CPL_TYPE_DOUBLE);
    /* Compute the fwhms */
    sofi_spc_arc_config.fwhm_good = 0;
    for (i=0; i<cpl_apertures_get_size(arcs); i++) {
        cpl_table_set_double(*arcs_fwhm, "POS", i,
                cpl_apertures_get_centroid_x(arcs, i+1));
        cpl_table_set_double(*arcs_fwhm, "FLUX", i,
                cpl_apertures_get_flux(arcs, i+1));
        if (cpl_image_get_fwhm(in_corr, cpl_apertures_get_centroid_x(arcs, i+1),
                512, &fwhm_x, &fwhm_y) != CPL_ERROR_NONE) {
            cpl_msg_warning(cpl_func, "Cannot get the FWHM");
            cpl_error_reset();
        }
        cpl_table_set_double(*arcs_fwhm, "FWHM", i, fwhm_x);
        if (fwhm_x > 0) sofi_spc_arc_config.fwhm_good ++;
    }
    cpl_apertures_delete(arcs);
    
    /* Compute the median of the good fwhms */
    if (sofi_spc_arc_config.fwhm_good > 0) {
        fwhm_valid = cpl_vector_new(sofi_spc_arc_config.fwhm_good);
        sofi_spc_arc_config.fwhm_good = 0;
        for (i=0; i<cpl_table_get_nrow(*arcs_fwhm); i++) {
            fwhm_x = cpl_table_get_double(*arcs_fwhm, "FWHM", i, NULL);
            if (fwhm_x > 0) {
                cpl_vector_set(fwhm_valid, sofi_spc_arc_config.fwhm_good, 
                        fwhm_x);
                sofi_spc_arc_config.fwhm_good ++;
            }
        }
        sofi_spc_arc_config.fwhm_med = cpl_vector_get_median_const(fwhm_valid);
        cpl_vector_delete(fwhm_valid);
    } else sofi_spc_arc_config.fwhm_med = 0.0;
    
    /* Wavelength calibration */

    /* Compute the slit_width */
    if ((slit_width = sofi_get_slitwidth(in_name)) == -1) {
        cpl_msg_error(cpl_func, "cannot get the slit width");
        cpl_polynomial_delete(distor_poly);
        cpl_image_delete(in_corr);
        cpl_table_delete(*arcs_fwhm);
        *arcs_fwhm = NULL;
        return NULL;
    }

    /* First estimation */
    cpl_msg_info(cpl_func, "Get the dispersion estimation");
    cpl_msg_indent_more();
    if ((phdisprel = sofi_get_disprel_estimate(in_name, 1)) == NULL) {
        cpl_msg_error(cpl_func, "cannot estimate the dispersion relation");
        cpl_polynomial_delete(distor_poly);
        cpl_image_delete(in_corr);
        cpl_table_delete(*arcs_fwhm);
        *arcs_fwhm = NULL;
        cpl_msg_indent_less();
        return NULL;
    }
    power[0] = 0; power[1] = 1;
    a = cpl_polynomial_get_coeff(phdisprel, power);
    b = cpl_polynomial_get_coeff(phdisprel, power + 1);
    cpl_msg_info(cpl_func, "f(x)=%g + %g*x", a, b);
    cpl_msg_indent_less();

    /* Compute the dispersion relation */
    cpl_msg_info(cpl_func, "Compute the dispersion relation");
    cpl_msg_indent_more();
    if ((disprel = sofi_wavelength_engine(in_corr, lines_table, NULL, xe, ne, 
                    phdisprel, slit_width, sofi_spc_arc_config.degree, 
                    sofi_spc_arc_config.wl_err, 
                    sofi_spc_arc_config.wl_nsamples, 
                    sofi_spc_arc_config.ppm, sofi_spc_arc_config.display, 
                    &xc)) == NULL) {
        cpl_msg_error(cpl_func, "cannot compute the dispersion relation");
        cpl_polynomial_delete(distor_poly);
        cpl_image_delete(in_corr);
        cpl_polynomial_delete(phdisprel);
        cpl_table_delete(*arcs_fwhm);
        *arcs_fwhm = NULL;
        cpl_msg_indent_less();
        return NULL;
    }
    if (sofi_spc_arc_config.out_corr) {
        *out_corr = in_corr;
        in_corr = NULL;
    } else {
        cpl_image_delete(in_corr);
    }

    /* Plot if requested */
    if (sofi_spc_arc_config.display && sofi_spc_arc_config.lines != NULL &&
        sofi_spc_arc_config.lines[0] != (char)0) {

        cpl_bivector * lines_biv
            = cpl_bivector_read(sofi_spc_arc_config.lines);
        const int nlines = cpl_bivector_get_size(lines_biv);

        if (nlines > sofi_spc_arc_config.degree) {

            /* Compute the solution from the passed associated lines */
            cpl_matrix * xpos
                = cpl_matrix_wrap(1, nlines, cpl_bivector_get_x_data(lines_biv));
            cpl_polynomial * poly_sol = cpl_polynomial_new(1);

            if (!cpl_polynomial_fit(poly_sol, xpos, NULL,
                                    cpl_bivector_get_y(lines_biv), NULL,
                                    CPL_FALSE, NULL,
                                    &(sofi_spc_arc_config.degree))) {
                irplib_wlxcorr_plot_solution(phdisprel, disprel, poly_sol, 1,
                                             1024);
            }
            (void)cpl_matrix_unwrap(xpos);
            cpl_polynomial_delete(poly_sol);
        }
        cpl_bivector_delete(lines_biv);
    }
    cpl_polynomial_delete(phdisprel);
    cpl_msg_info(cpl_func, "Cross correlation factor: %g", xc);
    power[0] = 0; power[1] = 1; power[2] = 2; power[3] = 3; power[4] = 4;
    a = b = c = d = e = 0.0;
    degree = cpl_polynomial_get_degree(disprel);
    a = cpl_polynomial_get_coeff(disprel, power);
    if (degree > 0) b = cpl_polynomial_get_coeff(disprel, power + 1);
    if (degree > 1) c = cpl_polynomial_get_coeff(disprel, power + 2);
    if (degree > 2) d = cpl_polynomial_get_coeff(disprel, power + 3);
    if (degree > 3) e = cpl_polynomial_get_coeff(disprel, power + 4);
    cpl_polynomial_delete(disprel);
    if (degree == 0)
        cpl_msg_info(cpl_func, "f(x)=%g", a);
    if (degree == 1)
        cpl_msg_info(cpl_func, "f(x)=%g + %g*x", a, b);
    if (degree == 2)
        cpl_msg_info(cpl_func, "f(x)=%g + %g*x + %g*x^2", a, b, c);
    if (degree == 3)
        cpl_msg_info(cpl_func, "f(x)=%g + %g*x + %g*x^2 + %g*x^3", a, b, c, d);
    if (degree == 4)
        cpl_msg_info(cpl_func, "f(x)=%g + %g*x + %g*x^2 + %g*x^3 + %g*x^4", 
                a, b, c, d, e);
    cpl_msg_indent_less();
    
    /* Create the output table and store the relevant information */
    out_tab = cpl_table_new(6);
    
    /* Fill distortion polynomial */
    cpl_table_new_column(out_tab, "Degree_of_x", CPL_TYPE_INT);
    cpl_table_new_column(out_tab, "Degree_of_y", CPL_TYPE_INT);
    cpl_table_new_column(out_tab, "poly2d_coef", CPL_TYPE_DOUBLE);
    power[0] = 0; power[1] = 0;
    cpl_table_set_int(out_tab, "Degree_of_x", 0, power[0]);
    cpl_table_set_int(out_tab, "Degree_of_y", 0, power[1]);
    cpl_table_set_double(out_tab, "poly2d_coef", 0,
            cpl_polynomial_get_coeff(distor_poly, power));
    power[0] = 1; power[1] = 0;
    cpl_table_set_int(out_tab, "Degree_of_x", 1, power[0]);
    cpl_table_set_int(out_tab, "Degree_of_y", 1, power[1]);
    cpl_table_set_double(out_tab, "poly2d_coef", 1,
            cpl_polynomial_get_coeff(distor_poly, power));
    power[0] = 0; power[1] = 1;
    cpl_table_set_int(out_tab, "Degree_of_x", 2, power[0]);
    cpl_table_set_int(out_tab, "Degree_of_y", 2, power[1]);
    cpl_table_set_double(out_tab, "poly2d_coef", 2,
            cpl_polynomial_get_coeff(distor_poly, power));
    power[0] = 1; power[1] = 1;
    cpl_table_set_int(out_tab, "Degree_of_x", 3, power[0]);
    cpl_table_set_int(out_tab, "Degree_of_y", 3, power[1]);
    cpl_table_set_double(out_tab, "poly2d_coef", 3,
            cpl_polynomial_get_coeff(distor_poly, power));
    power[0] = 2; power[1] = 0;
    cpl_table_set_int(out_tab, "Degree_of_x", 4, power[0]);
    cpl_table_set_int(out_tab, "Degree_of_y", 4, power[1]);
    cpl_table_set_double(out_tab, "poly2d_coef", 4,
            cpl_polynomial_get_coeff(distor_poly, power));
    power[0] = 0; power[1] = 2;
    cpl_table_set_int(out_tab, "Degree_of_x", 5, power[0]);
    cpl_table_set_int(out_tab, "Degree_of_y", 5, power[1]);
    cpl_table_set_double(out_tab, "poly2d_coef", 5,
            cpl_polynomial_get_coeff(distor_poly, power));

    /* Fill dispersion relation */
    cpl_table_new_column(out_tab, "WL_coefficients", CPL_TYPE_DOUBLE);
    cpl_table_set_double(out_tab, "WL_coefficients", 0, a);
    cpl_table_set_double(out_tab, "WL_coefficients", 1, b);
    cpl_table_set_double(out_tab, "WL_coefficients", 2, c);
    cpl_table_set_double(out_tab, "WL_coefficients", 3, d);
    cpl_table_set_double(out_tab, "WL_coefficients", 4, e);
    cpl_table_set_double(out_tab, "WL_coefficients", 5, 0.0);
    sofi_spc_arc_config.disprel_cc     = xc;

    /* Free and return */
    cpl_polynomial_delete(distor_poly);
    return out_tab;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the recipe products on disk
  @param    set_tot the input frame set
  @param    tab     table to save
  @param    fwhms   table with arcs fwhm values
  @param    corr    distortion corrected image
  @param    lines_table catalog
  @param    set     the frames from the current setting
  @param    parlist     the input list of parameters
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code sofi_spc_arc_save(cpl_frameset            * set_tot,
                                        const cpl_table         * tab,
                                        const cpl_table         * fwhms,
                                        const cpl_image         * corr,
                                        const char              * lines_table,
                                        const cpl_frameset      * set,
                                        const cpl_parameterlist * parlist)
{
    cpl_propertylist    *   plist;
    cpl_propertylist    *   qclist;
    cpl_propertylist    *   paflist;
    const cpl_frame     *   ref_frame;
    const char          *   sval;
    char                    qc_str[128];
    char                *   filename;
    int                     i;
    
    /* Get the QC params in qclist */
    qclist = cpl_propertylist_new();
    
    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return cpl_error_set_where(cpl_func);
    }
    /* Test the status */
    if (cpl_error_get_code()) {
        cpl_propertylist_delete(qclist);
        cpl_propertylist_delete(plist);
        return cpl_error_set_where(cpl_func);
    }
    sval = sofi_pfits_get_filter(plist);
    if (cpl_error_get_code()) cpl_error_reset();
    else cpl_propertylist_append_string(qclist, "ESO QC FILTER OBS", sval);
    cpl_propertylist_delete(plist);
    cpl_propertylist_append_string(qclist, "ESO QC LAMP", lines_table);
    cpl_propertylist_append_double(qclist, "ESO QC DISP XCORR",
            sofi_spc_arc_config.disprel_cc);
    cpl_propertylist_append_double(qclist, "ESO QC DISPCO1",
            cpl_table_get_double(tab, "WL_coefficients", 0, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DISPCO2",
            cpl_table_get_double(tab, "WL_coefficients", 1, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DISPCO3",
            cpl_table_get_double(tab, "WL_coefficients", 2, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DISPCO4",
            cpl_table_get_double(tab, "WL_coefficients", 3, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DISPCO5",
            cpl_table_get_double(tab, "WL_coefficients", 4, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DIST1",
            cpl_table_get_double(tab, "poly2d_coef", 0, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DISTX",
            cpl_table_get_double(tab, "poly2d_coef", 1, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DISTY",
            cpl_table_get_double(tab, "poly2d_coef", 2, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DISTXY",
            cpl_table_get_double(tab, "poly2d_coef", 3, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DISTXX",
            cpl_table_get_double(tab, "poly2d_coef", 4, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DISTYY",
            cpl_table_get_double(tab, "poly2d_coef", 5, NULL));
    cpl_propertylist_append_int(qclist, "ESO QC SATUR NBPIX",
            sofi_spc_arc_config.nb_saturated);
    cpl_propertylist_append_double(qclist, "ESO QC WLEN",
        (double)(cpl_table_get_double(tab, "WL_coefficients", 0, NULL) +
             512.0*cpl_table_get_double(tab, "WL_coefficients", 1, NULL) +
         512.0*512.0*cpl_table_get_double(tab, "WL_coefficients", 2, NULL) +
     512.0*512.0*512.0*cpl_table_get_double(tab, "WL_coefficients", 3, NULL) +
512.0*512.0*512.0*512.0*cpl_table_get_double(tab, "WL_coefficients", 4, NULL)));

    if (fwhms) {
        cpl_propertylist_append_int(qclist, "ESO QC ARCS NUM",
                cpl_table_get_nrow(fwhms));
        for(i=0; i<cpl_table_get_nrow(fwhms); i++) {
            sprintf(qc_str, "ESO QC ARCS%d XPOS", i+1);
            cpl_propertylist_append_double(qclist, qc_str,
                    cpl_table_get_double(fwhms, "POS", i, NULL));
            sprintf(qc_str, "ESO QC ARCS%d FWHM", i+1);
            cpl_propertylist_append_double(qclist, qc_str,
                    cpl_table_get_double(fwhms, "FWHM", i, NULL));
            sprintf(qc_str, "ESO QC ARCS%d FLUX", i+1);
            cpl_propertylist_append_double(qclist, qc_str,
                    cpl_table_get_double(fwhms, "FLUX", i, NULL));
        }
        cpl_propertylist_append_int(qclist, "ESO QC ARCS NUMGOOD",
                sofi_spc_arc_config.fwhm_good);
        cpl_propertylist_append_double(qclist, "ESO QC FWHM MED",
                sofi_spc_arc_config.fwhm_med);
    }

    /* Write the arc table  */
    filename = cpl_sprintf("sofi_spc_arc_set%d_frame%d.fits",
            sofi_spc_arc_config.set_nb, sofi_spc_arc_config.frame_nb);
    irplib_dfs_save_table(set_tot,
                parlist,
                set,
                tab,
                NULL,
                "sofi_spc_arc",
                SOFI_SPC_ARC_RES,
                qclist,
                NULL,
                PACKAGE "/" PACKAGE_VERSION,
                filename);
    cpl_free(filename);

    /* Write the corrected image  */
    if (corr) {
        filename = cpl_sprintf("sofi_spc_arc_set%d_frame%d_corr.fits",
                sofi_spc_arc_config.set_nb, sofi_spc_arc_config.frame_nb);
        irplib_dfs_save_image(set_tot,
                parlist,
                set,
                corr,
                CPL_BPP_IEEE_FLOAT,
                "sofi_spc_arc",
                SOFI_SPC_ARC_CORR,
                qclist,
                NULL,
                PACKAGE "/" PACKAGE_VERSION,
                filename);
        cpl_free(filename);
    }

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);

    /* Get FITS header from reference file */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return cpl_error_set_where(cpl_func);
    }

    /* Get the keywords for the paf file */
    paflist = cpl_propertylist_new();
    cpl_propertylist_copy_property_regexp(paflist, plist,
        "^(ARCFILE|MJD-OBS|INSTRUME|ESO TPL ID|ESO TPL NEXP|ESO DPR CATG|"
        "ESO DPR TECH|ESO DPR TYPE|DATE-OBS|ESO INS GRAT NAME|"
        "ESO INS GRAT WLEN|ESO INS GRAT ORDER|ESO INS MODE|"
        "ESO INS OPTI1 ID)$", 0);
    cpl_propertylist_delete(plist);

    /* Copy the QC in paflist */
    cpl_propertylist_copy_property_regexp(paflist, qclist, ".", 0);
    cpl_propertylist_delete(qclist);

    /* Save the PAF file */
    filename = cpl_sprintf("sofi_spc_arc_set%d_frame%d.paf",
            sofi_spc_arc_config.set_nb, sofi_spc_arc_config.frame_nb);
    cpl_dfs_save_paf("SOFI",
            "sofi_spc_arc",
            paflist,
            filename);
    cpl_free(filename);
    cpl_propertylist_delete(paflist);

    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Comparison function to identify different settings
  @param    frame1  first frame 
  @param    frame2  second frame 
  @return   0 if frame1!=frame2, 1 if frame1==frame2, -1 in error case
 */
/*----------------------------------------------------------------------------*/
static int sofi_spc_arc_compare(
        const cpl_frame *   frame1, 
        const cpl_frame *   frame2) 
{
    int                     comparison;
    cpl_propertylist    *   plist1;
    cpl_propertylist    *   plist2;
    const char          *   sval1,
                        *   sval2;

    /* Test entries */
    if (frame1==NULL || frame2==NULL) return CPL_ERROR_UNSPECIFIED;

    /* Get property lists */
    if ((plist1=cpl_propertylist_load(cpl_frame_get_filename(frame1),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        return CPL_ERROR_UNSPECIFIED;
    }
    if ((plist2=cpl_propertylist_load(cpl_frame_get_filename(frame2),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(plist1);
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Test status */
    if (cpl_error_get_code()) {
        cpl_propertylist_delete(plist1);
        cpl_propertylist_delete(plist2);
        return CPL_ERROR_UNSPECIFIED;
    }

    comparison = 1;

    /* Compare the slit used */
    sval1 = sofi_pfits_get_opti1_id(plist1);
    sval2 = sofi_pfits_get_opti1_id(plist2);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "cannot get the slit used");
        cpl_propertylist_delete(plist1);
        cpl_propertylist_delete(plist2);
        return CPL_ERROR_UNSPECIFIED;
    }
    if (strcmp(sval1, sval2)) comparison = 0;

    /* Compare the mode */
    sval1 = sofi_pfits_get_mode(plist1);
    sval2 = sofi_pfits_get_mode(plist2);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "cannot get the mode");
        cpl_propertylist_delete(plist1);
        cpl_propertylist_delete(plist2);
        return CPL_ERROR_UNSPECIFIED;
    }
    if (strcmp(sval1, sval2)) comparison = 0;

    cpl_propertylist_delete(plist1);
    cpl_propertylist_delete(plist2);
    return comparison;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    identify the activated lamps in a frames list   
  @param    fset    list of frames
  @return   array: array[i]: 0 if no activated lamps, 1 if xenon lamp is 
  activated, 2 if neon lamp is activated, 3 if both lamps are activated
 */
/*----------------------------------------------------------------------------*/
static int * sofi_spc_arc_find_lamps(const cpl_frameset * fset)
{
    const int   nframes = cpl_frameset_get_size(fset);
    int       * lamps;
    int         i;

    cpl_ensure(nframes > 0, CPL_ERROR_UNSPECIFIED, NULL); /* A bug */
  
    /* Allocate the output array */
    lamps = (int*)cpl_malloc(nframes * sizeof(int));

    for (i=0; i<nframes; i++) {
        const cpl_frame  * cur_frame = cpl_frameset_get_position_const(fset, i);
        cpl_propertylist * plist
            = cpl_propertylist_load(cpl_frame_get_filename(cur_frame), 0);
        const int xenon = sofi_is_xenon_lamp_active(plist);
        const int neon  = sofi_is_neon_lamp_active(plist);

        cpl_propertylist_delete(plist);
        if ((neon != 0 && neon != 1) || (xenon != 0 && xenon != 1)) {
            cpl_free(lamps);
            (void)cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                                        "Cannot check which lamp is on for "
                                        "frame %d", i+1);
            return NULL;
        }
        if (xenon == 1) {
            if (neon == 1)         lamps[i] = 3;
            else if (neon == 0)    lamps[i] = 1;
        } else if (xenon == 0 ) {
            if (neon == 1)         lamps[i] = 2;
            else if (neon == 0)    lamps[i] = 0;
        }
    }

    return lamps;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out if the neon lamp was active
  @param    plist   the property list
  @return   1 if the lamp is active, 0 if not, -1 in error case
 */
/*----------------------------------------------------------------------------*/
static int sofi_is_neon_lamp_active(const cpl_propertylist * plist)
{
    int             status;
    const char  *   lamp;

    /* Test error */
    if (cpl_error_get_code()) return CPL_ERROR_UNSPECIFIED;

    if ((lamp=sofi_pfits_get_lamp2_name(plist)) == NULL) {
        cpl_error_reset();
        return 0;
    }
    if (!strcmp(lamp, "Neon")) {
        status = sofi_pfits_get_lamp2_status(plist);
        if (cpl_error_get_code()) {
            cpl_error_reset();
            return 0;
        }
        if (status == 1) return 1;
    }
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out if the Xenon lamp was active
  @param    plist   the property list
  @return   1 if the lamp is active, 0 if not, -1 in error case
 */
/*----------------------------------------------------------------------------*/
static int sofi_is_xenon_lamp_active(const cpl_propertylist * plist)
{
    int             status;
    const char  *   lamp;

    /* Test error */
    if (cpl_error_get_code()) return CPL_ERROR_UNSPECIFIED;

    if ((lamp=sofi_pfits_get_lamp1_name(plist)) == NULL) {
        cpl_error_reset();
        return 0;
    }
    if (!strcmp(lamp, "Xenon")) {
        status = sofi_pfits_get_lamp1_status(plist);
        if (cpl_error_get_code()) {
            cpl_error_reset();
            return 0;
        }
        if (status == 1) return 1;
    }
    return 0;
}
