/* $Id: sofi_util_stdstars.c,v 1.10 2013-03-12 08:04:52 llundin Exp $
 *
 * This file is part of the SOFI Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:04:52 $
 * $Revision: 1.10 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <math.h>
#include <cpl.h>

#include "irplib_utils.h"
#include "irplib_stdstar.h"

#include "sofi_utils.h"
#include "sofi_pfits.h"
#include "sofi_dfs.h"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int sofi_util_stdstars_create(cpl_plugin *) ;
static int sofi_util_stdstars_exec(cpl_plugin *) ;
static int sofi_util_stdstars_destroy(cpl_plugin *) ;
static int sofi_util_stdstars(cpl_frameset *) ;
static cpl_table * sofi_util_stdstars_convert(const char *) ; 

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static char sofi_util_stdstars_description[] = 
"sofi_util_stdstars -- SOFI standard stars catalog creation.\n"
"The files listed in the Set Of Frames (sof-file) must be tagged:\n"
"raw-file.fits "SOFI_UTIL_STDSTARS_RAW"\n" ;

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe)) ;
    cpl_plugin  *   plugin = &recipe->interface ;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    SOFI_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "sofi_util_stdstars",
                    "Standard stars catalog creation",
                    sofi_util_stdstars_description,
                    "Yves Jung",
                    "yjung@eso.org",
                    sofi_get_license(),
                    sofi_util_stdstars_create,
                    sofi_util_stdstars_exec,
                    sofi_util_stdstars_destroy) ;

    cpl_pluginlist_append(list, plugin) ;
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
static int sofi_util_stdstars_create(cpl_plugin * plugin)
{
    cpl_recipe      * recipe ;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin ;
    else return CPL_ERROR_UNSPECIFIED;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new() ;

    /* Return */
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int sofi_util_stdstars_exec(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe ;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin ;
    else return CPL_ERROR_UNSPECIFIED;

    return sofi_util_stdstars(recipe->frames) ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int sofi_util_stdstars_destroy(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe ;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin ;
    else return CPL_ERROR_UNSPECIFIED;

    cpl_parameterlist_delete(recipe->parameters) ;
    return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int sofi_util_stdstars(
        cpl_frameset        *   framelist)
{
    cpl_frameset    *   rawframes ;
    
    /* Identify the RAW and CALIB frames in the input frameset */
    if (sofi_dfs_set_groups(framelist)) {
        cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames") ;
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Retrieve raw frames */
    if ((rawframes = sofi_extract_frameset(framelist,
                    SOFI_UTIL_STDSTARS_RAW)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot find raw frames in the input list") ;
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Write the catalog */
    if (irplib_stdstar_write_catalogs(framelist,
                rawframes, 
                "sofi_util_stdstars",
                SOFI_UTIL_STDSTARS_RES, 
                NULL,
                PACKAGE "/" PACKAGE_VERSION,
                "SOFI",
                sofi_util_stdstars_convert)) {
        cpl_msg_error(cpl_func, "Cannot write the catalogs") ;
        cpl_frameset_delete(rawframes) ;
        return CPL_ERROR_UNSPECIFIED;
    }
    cpl_frameset_delete(rawframes) ;
    return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Convert the ASCII catalog to the table
  @param    filename    the ascii file
  @return   the newly allocated table or NULL in error case

  The returned table must be deallocated with cpl_table_delete().

  The inputcatalogs are ASCII files respecting this format:

  Star_name RA DEC Spec_type MAG_J MAG_H MAG_K MAG_KS MAG_L MAG_M MAG_LP MAG_MP

  Example:
  AS01-0 13.7912 0.720278 -- 10.716 10.507 10.470 99 99 99 99 99
  AS03-0 16.09 4.2275 -- 12.606 12.729 12.827 99 99 99 99 99
  AS04-1 28.6808 0.733056 -- 12.371 12.033 11.962 99 99 99 99 99
  ... 

  99 is the value for unknown magnitudes.
  
 */
/*----------------------------------------------------------------------------*/
static cpl_table * sofi_util_stdstars_convert(const char * filename) 
{
    cpl_table   *   out ;
    int             nfilters ;
    const char  *   filters[8];
    double          mags[8] ;
    int             nbentries ;
    FILE        *   in ;
    char            line[1024];
    double          ra, dec ;
    char            sname[512];
    char            stype[512];
    int             i ;
    
    /* Check entries */
    if (filename == NULL) {
        (void)cpl_error_set(cpl_func, CPL_ERROR_NULL_INPUT);
        return NULL;
    }

    /* Initialise */
    nfilters = 8 ;
    filters[0] = "J" ;
    filters[1] = "H" ;
    filters[2] = "K" ;
    filters[3] = "Ks" ;
    filters[4] = "L" ;
    filters[5] = "M" ;
    filters[6] = "Lp" ;
    filters[7] = "Mp" ;
   
    /* Get the number of lines */
    nbentries = 0 ;
    if ((in = fopen(filename, "r")) == NULL) {
        (void)cpl_error_set(cpl_func, CPL_ERROR_FILE_IO);
        return NULL ;
    }
    while (fgets(line, 1024, in) != NULL) {
        if (line[0] != '#') nbentries ++ ;
    }
    rewind(in); /* FIXME: Avoid this.... */

   /* Create the table */
    out = cpl_table_new(nbentries);
    cpl_table_new_column(out, IRPLIB_STDSTAR_STAR_COL, CPL_TYPE_STRING);
    cpl_table_new_column(out, IRPLIB_STDSTAR_TYPE_COL, CPL_TYPE_STRING);
    cpl_table_new_column(out, IRPLIB_STDSTAR_RA_COL, CPL_TYPE_DOUBLE);
    cpl_table_new_column(out, IRPLIB_STDSTAR_DEC_COL, CPL_TYPE_DOUBLE);
    for (i=0 ; i<nfilters ; i++)
        cpl_table_new_column(out, filters[i], CPL_TYPE_DOUBLE);

    /* Parse the file */
    nbentries = 0 ;
    while (fgets(line, 1024, in) != NULL) {
        if (line[0] != '#') {
            if (sscanf(line,
                       "%511s %lg %lg %511s %lg %lg %lg %lg %lg %lg %lg %lg",
                        sname, &ra, &dec, stype, &(mags[0]), &(mags[1]), 
                        &(mags[2]), &(mags[3]), &(mags[4]), &(mags[5]), 
                        &(mags[6]), &(mags[7])) != 12) {
                fclose(in);
                cpl_table_delete(out) ;
                (void)cpl_error_set(cpl_func, CPL_ERROR_BAD_FILE_FORMAT);
                return NULL ;
            }
            cpl_table_set_string(out, IRPLIB_STDSTAR_STAR_COL, nbentries,sname);
            cpl_table_set_string(out, IRPLIB_STDSTAR_TYPE_COL, nbentries,stype);
            cpl_table_set_double(out, IRPLIB_STDSTAR_RA_COL, nbentries, ra);
            cpl_table_set_double(out, IRPLIB_STDSTAR_DEC_COL, nbentries, dec);
            for (i=0 ; i<nfilters ; i++)
                cpl_table_set_double(out, filters[i], nbentries, mags[i]);
            nbentries ++ ;
        }
    }
    fclose(in) ;

    if (cpl_table_get_nrow(out) == 0) {
        cpl_table_delete(out);
        (void)cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                    "Empty catalogue in '%s'", filename);
        return NULL ;
    }
                
    return out ;
}

