/* $Id: sofi_pfits.c,v 1.6 2013-03-12 08:04:54 llundin Exp $
 *
 * This file is part of the SOFI Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:04:54 $
 * $Revision: 1.6 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
   								Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <cpl.h>

#include "sofi_pfits.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup sofi_pfits     FITS header protected access
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*-----------------------------------------------------------------------------
  							Function codes
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    find out airmass start
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double sofi_pfits_get_airmass_start(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO TEL AIRM START") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out airmass end
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double sofi_pfits_get_airmass_end(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO TEL AIRM END") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the arcfile   
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_arcfile(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, "ARCFILE");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the chip name
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_chip(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, 
            "ESO DET CHIP NAME");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the cumulative offset in X
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double sofi_pfits_get_cumoffsetx(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO SEQ CUMOFFSETX") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the cumulative offset in Y
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double sofi_pfits_get_cumoffsety(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO SEQ CUMOFFSETY") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the current exposure number
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int sofi_pfits_get_current_exp_nb(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_int(plist, "ESO TPL EXPNO") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the date of the file creation
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_date(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, "DATE") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the date of observation 
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_date_obs(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, "DATE-OBS") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the DEC
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double sofi_pfits_get_dec(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "DEC") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the frame type
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_frame_type(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, 
            "ESO DET FRAM TYPE") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the DIT value 
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double sofi_pfits_get_dit(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO DET DIT")  ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the data category
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_dpr_catg(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, "ESO DPR CATG") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the data technique
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_dpr_tech(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, "ESO DPR TECH") ;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out the data type
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_dpr_type(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, "ESO DPR TYPE") ;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out which wave band is active in short wavelength
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_filter(const cpl_propertylist * plist)
{
    const char  *   val ;

    val = cpl_propertylist_get_string(plist, "ESO INS FILT1 ID") ;
    if (val == NULL) return NULL ;
    /* If FILT1 is not open, return its value */
    if (strcmp(val, "open")) return val ;
    /* FILT1 is open, return value from FILT2 */
    val = cpl_propertylist_get_string(plist, "ESO INS FILT2 ID") ;
    if (val == NULL) return NULL ;
    if (strcmp(val, "open")) return val ;
    return NULL ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the humidity level
  @param    plist       property list to read from
  @return   the requested value 
 */
/*----------------------------------------------------------------------------*/
double sofi_pfits_get_humidity_level(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO TEL AMBI RHUM") ;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out the instrument
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_instrument(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, "INSTRUME") ;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out the lamp1 name
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_lamp1_name(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist,
            "ESO INS LAMP1 NAME") ;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out the lamp1 status
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
int sofi_pfits_get_lamp1_status(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_bool(plist, "ESO INS LAMP1 ST") ;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out the lamp2 name
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_lamp2_name(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist,
            "ESO INS LAMP2 NAME") ;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out the lamp2 status
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
int sofi_pfits_get_lamp2_status(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_bool(plist, "ESO INS LAMP2 ST") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the MJD-OBS
  @param    plist       property list to read from
  @return   the requested value 
 */
/*----------------------------------------------------------------------------*/
double sofi_pfits_get_mjdobs(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "MJD-OBS") ;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out the instrument mode
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_mode(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, "ESO INS MODE") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the NAXIS2 keyword
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int sofi_pfits_get_naxis2(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_int(plist, "NAXIS2") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the NDIT keyword
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int sofi_pfits_get_ndit(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_int(plist, "ESO DET NDIT") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the number of exposures
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int sofi_pfits_get_numbexp(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_int(plist, "ESO TPL NEXP") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the OBS ID
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int sofi_pfits_get_obs_id(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_int(plist, "ESO OBS ID") ;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out the target name
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_targ_name(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, 
            "ESO OBS TARG NAME") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the OPTI1.ID key
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_opti1_id(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist,
            "ESO INS OPTI1 ID") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the original file name
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_origfile(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, "ORIGFILE") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the pixel scale
  @param    plist       property list to read from
  @return   the requested value 
 */
/*----------------------------------------------------------------------------*/
double sofi_pfits_get_pixscale(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "ESO INS PIXSCALE") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the RA
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double sofi_pfits_get_ra(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_double(plist, "RA") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the resolution
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_resolution(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, 
            "ESO INS OPTI2 NAME") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the read out mode
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int sofi_pfits_get_rom(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_int(plist, "ESO DET NCORRS") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the read out mode name
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */         
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_romode_name(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist,
            "ESO DET MODE NAME");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the read out mode name
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */         
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_rom_name(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist,
            "ESO DET NCORRS NAME");
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the read speed
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int sofi_pfits_get_rspeed(const cpl_propertylist * plist)
{
    return cpl_propertylist_get_int(plist, "ESO DET RSPEED") ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the template ID
  @param    plist       property list to read from
  @return   pointer to statically allocated character string
 */
/*----------------------------------------------------------------------------*/
const char * sofi_pfits_get_templateid(const cpl_propertylist * plist)
{
    return (const char *) cpl_propertylist_get_string(plist, "ESO TPL ID") ;
}

/**@}*/
